from documenteer.conf.guide import *

autodoc_pydantic_field_swap_name_and_alias = True
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_settings_show_config_summary = False
autodoc_pydantic_settings_show_json = False



"""pytest fixtures for neophile testing."""

from __future__ import annotations

from collections.abc import AsyncIterator, Iterator
from unittest.mock import Mock, patch

import pytest
import pytest_asyncio
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
)
from git import PushInfo, Remote
from httpx import AsyncClient

__all__ = [
    "client",
    "github_key",
    "mock_push",
]


@pytest_asyncio.fixture
async def client() -> AsyncIterator[AsyncClient]:
    """Return an `httpx.AsyncClient` for testing."""
    async with AsyncClient() as client:
        yield client


@pytest.fixture(scope="session")
def github_key() -> str:
    """RSA private key for mock GitHub API."""
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )
    pem = private_key.private_bytes(
        Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
    )
    return pem.decode()


@pytest.fixture
def mock_push() -> Iterator[Mock]:
    """Mock out `git.Remote.push`.

    The mock will always return success with a status indicating that a new
    remote head was created.
    """
    with patch.object(Remote, "push") as mock:
        remote = Mock(spec=Remote)
        mock.return_value = [PushInfo(PushInfo.NEW_HEAD, None, "", remote)]
        yield mock



"""Base class for dependency scanners."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from pathlib import Path

from ..models.dependencies import Dependency

__all__ = ["BaseScanner"]


class BaseScanner(ABC):
    """Base class for dependency scanners."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the scanner type."""

    @abstractmethod
    def scan(self, root: Path) -> Sequence[Dependency]:
        """Scan a source tree for dependencies.

        Parameters
        ----------
        root
            Root of the source tree.

        Returns
        -------
        list of Dependency
            A list of all discovered dependencies.
        """



"""Tests for the PythonFrozenUpdate class."""

from __future__ import annotations

import re
import shutil
from pathlib import Path

from neophile.update.python import PythonFrozenUpdate


def test_python_update(tmp_path: Path) -> None:
    data_path = Path(__file__).parent.parent / "data" / "python"
    main_path = tmp_path / "requirements" / "main.txt"
    shutil.copytree(str(data_path), str(tmp_path), dirs_exist_ok=True)

    with (data_path / "Makefile").open() as f:
        for line in f:
            match = re.match("NEW = (.*)", line)
            if match:
                new_hash = match.group(1)
    assert new_hash not in main_path.read_text()

    update = PythonFrozenUpdate(path=tmp_path / "requirements", applied=False)
    assert "Python" in update.description()
    update.apply()
    assert new_hash in main_path.read_text()


def test_python_update_applied(tmp_path: Path) -> None:
    data_path = Path(__file__).parent.parent / "data" / "python"
    main_path = tmp_path / "requirements" / "main.txt"
    shutil.copytree(str(data_path), str(tmp_path), dirs_exist_ok=True)
    main_data = main_path.read_text()

    update = PythonFrozenUpdate(path=tmp_path / "requirements", applied=True)
    update.apply()
    assert main_data == main_path.read_text()



"""Configuration for neophile."""

from __future__ import annotations

from pydantic import BaseSettings, Field, SecretStr

__all__ = ["Config"]


class Config(BaseSettings):
    """Configuration for neophile."""

    github_app_id: str = Field("", description="GitHub App ID")

    github_private_key: SecretStr = Field(
        SecretStr(""), description="GitHub App private key"
    )

    username: str = Field(
        "neophile-square[bot]",
        description=(
            "Username, used as the name for GitHub commits and as part of"
            " the email address if `commit_email` is not set. In the latter"
            " case, must be the same as the login attribute in the GitHub"
            " users API."
        ),
    )

    commit_email: str | None = Field(
        None,
        description=(
            "Email address to use for GitHub commits. If `None`, a GitHub"
            " bot email address will be constructed using `github_app_id`."
        ),
    )

    class Config:
        env_prefix = "neophile_"



"""Construct a GitHub PR for a set of changes."""

from __future__ import annotations

import logging
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar
from urllib.parse import ParseResult, urlencode, urlparse

from gidgethub import QueryError
from gidgethub.httpx import GitHubAPI
from git import PushInfo, Remote
from git.repo import Repo
from git.util import Actor
from httpx import AsyncClient
from safir.github import GitHubAppClientFactory

from .config import Config
from .exceptions import PushError
from .update.base import Update

__all__ = [
    "CommitMessage",
    "GitHubRepository",
    "PullRequester",
]

_GRAPHQL_PR_ID = """
query FindPrId($owner:String!, $repo:String!, $pr_number:Int!) {
  repository(owner: $owner, name: $repo) {
    pullRequest(number: $pr_number) {
      id
    }
  }
}
"""

_GRAPHQL_ENABLE_AUTO_MERGE = """
mutation EnableAutoMerge($pr_id:ID!) {
  enablePullRequestAutoMerge(input: {pullRequestId: $pr_id}) {
    actor {
      login
    }
  }
}
"""


@dataclass(frozen=True)
class CommitMessage:
    """A Git commit message."""

    changes: list[str]
    """Changes represented by this commit."""

    title: ClassVar[str] = "[neophile] Update dependencies"
    """Title of the commit message, currently fixed for all commits."""

    def __str__(self) -> str:
        return f"{self.title}\n\n{self.body}"

    @property
    def body(self) -> str:
        """Body of the commit message."""
        return "- " + "\n- ".join(self.changes) + "\n"


@dataclass(frozen=True)
class GitHubRepository:
    """An individual GitHub repository."""

    owner: str
    """The owner of the repository."""

    repo: str
    """The name of the repository."""


class PullRequester:
    """Create GitHub pull requests.

    Parameters
    ----------
    config
        neophile configuration.
    http_client
        HTTP client to use for requests.
    """

    def __init__(self, config: Config, http_client: AsyncClient) -> None:
        self._config = config
        self._factory = GitHubAppClientFactory(
            id=config.github_app_id,
            key=config.github_private_key.get_secret_value(),
            name="lsst-sqre/neophile",
            http_client=http_client,
        )

    async def make_pull_request(
        self, path: Path, changes: Sequence[Update]
    ) -> None:
        """Create or update a pull request for a list of changes.

        Parameters
        ----------
        path
            Path to the Git repository.
        changes
            The changes.

        Raises
        ------
        PushError
            Raised if pushing the branch to GitHub failed.
        """
        repo = Repo(str(path))
        github_repo = self._get_github_repo(repo)
        github = await self._factory.create_installation_client_for_repo(
            owner=github_repo.owner,
            repo=github_repo.repo,
        )
        if not github.oauth_token:
            # Mostly for mypy's benefit.
            raise RuntimeError("No OAuth token for installation client")
        default_branch = await self._get_github_default_branch(
            github, github_repo
        )
        pr_number = await self._get_pr(github, github_repo, default_branch)

        message = await self._commit_changes(repo, changes, github)
        self._push_branch(repo, github.oauth_token)
        if pr_number is not None:
            await self._update_pr(
                github=github,
                github_repo=github_repo,
                pr_number=pr_number,
                message=message,
            )
        else:
            await self._create_pr(
                github=github,
                repo=repo,
                github_repo=github_repo,
                base_branch=default_branch,
                message=message,
            )

    def _build_commit_message(
        self, changes: Sequence[Update]
    ) -> CommitMessage:
        """Build a commit message from a list of changes.

        Parameters
        ----------
        changes
            The changes.

        Returns
        -------
        CommitMessage
            Corresponding commit message.
        """
        descriptions = [change.description() for change in changes]
        return CommitMessage(changes=descriptions)

    async def _commit_changes(
        self, repo: Repo, changes: Sequence[Update], github: GitHubAPI
    ) -> CommitMessage:
        """Commit a set of changes to the repository.

        The changes will be committed on the current branch.

        Parameters
        ----------
        repo
            Local Git repository.
        changes
            Changes to apply and commit.
        github
            GitHub API client.

        Returns
        -------
        CommitMessage
            Commit message of the commit.
        """
        for change in changes:
            repo.index.add(str(change.path))
        message = self._build_commit_message(changes)
        username = self._config.username
        if self._config.commit_email:
            email = self._config.commit_email
        else:
            response = await github.getitem(
                "/users{/user}", url_vars={"user": username}
            )
            uid = str(response["id"])
            email = f"{uid}+{username}@users.noreply.github.com"
        actor = Actor(username, email)
        repo.index.commit(str(message), author=actor, committer=actor)
        return message

    async def _create_pr(
        self,
        *,
        github: GitHubAPI,
        repo: Repo,
        github_repo: GitHubRepository,
        base_branch: str,
        message: CommitMessage,
    ) -> None:
        """Create a new PR for the current branch.

        Parameters
        ----------
        github
            GitHub API client.
        repo
            Local Git repository.
        github_repo
            GitHub repository in which to create the pull request.
        base_branch
            Branch of the repository to use as the base for the PR.
        message
            Commit message to use for the pull request.
        """
        branch = repo.head.ref.name
        data = {
            "title": message.title,
            "body": message.body,
            "head": branch,
            "base": base_branch,
            "maintainer_can_modify": True,
            "draft": False,
        }
        response = await github.post(
            "/repos{/owner}{/repo}/pulls",
            url_vars={"owner": github_repo.owner, "repo": github_repo.repo},
            data=data,
        )
        pr_number = str(response["number"])
        await self._enable_auto_merge(github, github_repo, pr_number)

    async def _enable_auto_merge(
        self, github: GitHubAPI, github_repo: GitHubRepository, pr_number: str
    ) -> None:
        """Enable automerge for a PR.

        Failures are logged but do not raise an exception.

        Parameters
        ----------
        github
            GitHub API client.
        github_repo
            GitHub repository in which to create the pull request.
        pr_number
            Number of the PR in that repository to enable auto-merge for.
        """
        # Enabling auto-merge is only available via the GraphQL API with a
        # mutation.  To use that, we have to retrieve the GraphQL ID of the PR
        # we just created, and then send the mutation.
        try:
            response = await github.graphql(
                _GRAPHQL_PR_ID,
                owner=github_repo.owner,
                repo=github_repo.repo,
                pr_number=int(pr_number),
            )
            pr_id = response["repository"]["pullRequest"]["id"]
            response = await github.graphql(
                _GRAPHQL_ENABLE_AUTO_MERGE, pr_id=pr_id
            )
        except QueryError as e:
            msg = (
                f"cannot enable automerge for {github_repo.owner}/"
                f"{github_repo.repo}#{pr_number}: {e!s}"
            )
            logging.exception(msg)

    async def _get_github_default_branch(
        self, github: GitHubAPI, github_repo: GitHubRepository
    ) -> str:
        """Get the main branch of the repository.

        Uses the default branch name if it exists, else ``main``.

        Parameters
        ----------
        github
            GitHub API client.
        github_repo
            GitHub repository in which to create the pull request.

        Returns
        -------
        str
            Name of the main branch.
        """
        repo = await github.getitem(
            "/repos{/owner}{/repo}",
            url_vars={"owner": github_repo.owner, "repo": github_repo.repo},
        )
        return repo.get("default_branch", "main")

    def _get_github_repo(self, repo: Repo) -> GitHubRepository:
        """Get the GitHub repository.

        Done by parsing the URL of the origin remote.

        Returns
        -------
        GitHubRepository
            GitHub repository information.
        """
        url = next(repo.remotes.origin.urls)
        if "//" in url:
            parsed_url = urlparse(url)
        else:
            path = url.rsplit(":", 1)[-1]
            parsed_url = urlparse(f"https://github.com/{path}")
        owner, github_repo = parsed_url.path.lstrip("/").split("/")
        if github_repo.endswith(".git"):
            github_repo = github_repo[: -len(".git")]
        return GitHubRepository(owner=owner, repo=github_repo)

    async def _get_pr(
        self,
        github: GitHubAPI,
        github_repo: GitHubRepository,
        base_branch: str,
    ) -> str | None:
        """Get the pull request number of an existing neophile PR.

        Parameters
        ----------
        github
            GitHub API client.
        github_repo
            GitHub repository in which to search for a pull request.
        base_branch
            Base repository branch used to limit the search.

        Returns
        -------
        str or None
            PR number, or `None` if there is no open pull request from
            neophile.

        Notes
        -----
        The pull request is found by searching for all PRs in the open state
        whose branch is ``u/neophile``.
        """
        query = {
            "state": "open",
            "head": f"{github_repo.owner}:u/neophile",
            "base": base_branch,
        }

        prs = github.getiter(
            f"/repos{{/owner}}{{/repo}}/pulls?{urlencode(query)}",
            url_vars={"owner": github_repo.owner, "repo": github_repo.repo},
        )
        async for pr in prs:
            return str(pr["number"])
        return None

    def _get_remote_url(self, repo: Repo) -> ParseResult:
        """Get the parsed URL of the origin remote.

        The URL will be converted to https form.  https, ssh, and the SSH
        remote syntax used by Git are supported.

        Parameters
        ----------
        repo
            Local Git repository.

        Returns
        -------
        url
            Results of `~urllib.parse.urlparse` on the origin remote URL.
        """
        url = next(repo.remotes.origin.urls)
        if "//" in url:
            return urlparse(url)
        else:
            path = url.rsplit(":", 1)[-1]
            return urlparse(f"https://github.com/{path}")

    def _get_authenticated_remote(self, repo: Repo, token: str) -> str:
        """Construct an authenticated URL for use as a Git remote.

        When running in GitHub Actions, we need to push changes using the
        GitHub App installation token rather than the default GitHub Actions
        token so that GitHub Actions CI workflows will run.

        This is a separate method primarily for testing purposes.

        Parameters
        ----------
        repo
            Local Git repository. The ``origin`` remote's URL will be parsed
            to determine what GitHub repository this is a clone of.
        token
            GitHub App installation token.

        Returns
        -------
        str
            URL to use as an authenticated remote for Git pushes.
        """
        url = self._get_remote_url(repo)
        host = url.netloc.rsplit("@", 1)[-1]
        url = url._replace(scheme="https", netloc=f"neophile:{token}@{host}")
        return url.geturl()

    def _push_branch(self, repo: Repo, token: str) -> None:
        """Push the ``u/neophile`` branch to GitHub.

        The complication here is that when running in a GitHub Action, the
        default remote is authenticated using the GitHub Action token, which
        will not allow GitHub Action CI to run. We need to instead push to a
        remote using our GitHub App installation token.

        Parameters
        ----------
        repo
            Local Git repository.
        token
            GitHub App installation token.

        Raises
        ------
        PushError
            Raised if pushing the branch to GitHub failed.
        """
        branch = repo.head.ref.name
        url = self._get_authenticated_remote(repo, token)
        remote = Remote.add(repo, "tmp-neophile", url)
        try:
            push_info = remote.push(f"{branch}:{branch}", force=True)
            for result in push_info:
                if result.flags & PushInfo.ERROR:
                    msg = f"Pushing {branch} failed: {result.summary}"
                    raise PushError(msg)
        finally:
            Remote.remove(repo, "tmp-neophile")

    async def _update_pr(
        self,
        *,
        github: GitHubAPI,
        github_repo: GitHubRepository,
        pr_number: str,
        message: CommitMessage,
    ) -> None:
        """Update an existing PR with a new commit message.

        Parameters
        ----------
        github
            GitHub API client.
        github_repo
            GitHub repository in which to create the pull request.
        pr_number
            Number of the pull request to update.
        message
            Commit message to use for the pull request.
        """
        data = {
            "title": message.title,
            "body": message.body,
        }
        await github.patch(
            "/repos{/owner}{/repo}/pulls{/pull_number}",
            url_vars={
                "owner": github_repo.owner,
                "repo": github_repo.repo,
                "pull_number": pr_number,
            },
            data=data,
        )
        await self._enable_auto_merge(github, github_repo, pr_number)



"""Exceptions for neophile."""

from __future__ import annotations

__all__ = [
    "DependencyNotFoundError",
    "PushError",
    "UncommittedChangesError",
]


class DependencyNotFoundError(Exception):
    """The specified dependency was not found to update."""


class PushError(Exception):
    """Pushing a branch to GitHub failed."""


class UncommittedChangesError(Exception):
    """The repository contains uncommitted changes.

    This interferes with some types of dependency analysis, which rely on
    checking whether an action causes repository files to change.
    """



"""Tests for the GitHubInventory class."""

from __future__ import annotations

import pytest
import respx
from httpx import AsyncClient, Response

from neophile.inventory.github import GitHubInventory

from ..support.github import mock_github_tags


@pytest.mark.asyncio
async def test_inventory(
    client: AsyncClient, respx_mock: respx.Router
) -> None:
    tests = [
        {"tags": ["3.7.0", "3.8.0", "3.9.0", "3.8.1"], "latest": "3.9.0"},
        {"tags": ["v3.1.0", "v3.0.1", "v3.0.0", "v2.5.0"], "latest": "v3.1.0"},
        {"tags": ["4.3.20", "4.3.21-2"], "latest": "4.3.21-2"},
        {"tags": ["19.10b0", "19.3b0", "18.4a4"], "latest": "19.10b0"},
    ]

    for test in tests:
        mock_github_tags(respx_mock, "foo/bar", test["tags"])
        inventory = GitHubInventory(client)
        latest = await inventory.inventory("foo", "bar")
        assert latest == test["latest"]


@pytest.mark.asyncio
async def test_inventory_semantic(
    client: AsyncClient, respx_mock: respx.Router
) -> None:
    tags = ["1.19.0", "1.18.0", "1.15.1", "20171120-1"]

    mock_github_tags(respx_mock, "foo/bar", tags)
    inventory = GitHubInventory(client)
    latest = await inventory.inventory("foo", "bar")
    assert latest == "20171120-1"
    latest = await inventory.inventory("foo", "bar", semantic=True)
    assert latest == "1.19.0"


@pytest.mark.asyncio
async def test_inventory_missing(
    client: AsyncClient, respx_mock: respx.Router
) -> None:
    """Missing and empty version lists should return None."""
    mock_github_tags(respx_mock, "foo/bar", [])
    respx_mock.get("https://api.github.com/repos/foo/nonexistent/tags").mock(
        return_value=Response(404)
    )
    inventory = GitHubInventory(client)
    assert await inventory.inventory("foo", "bar") is None
    assert await inventory.inventory("foo", "nonexistent") is None



"""Analysis of a repository for needed Python updates."""

from __future__ import annotations

import logging
import subprocess
from pathlib import Path

from git.repo import Repo

from ..exceptions import UncommittedChangesError
from ..update.base import Update
from ..update.python import PythonFrozenUpdate
from .base import BaseAnalyzer

__all__ = ["PythonAnalyzer"]


class PythonAnalyzer(BaseAnalyzer):
    """Analyze a tree for needed Python frozen dependency updates.

    Parameters
    ----------
    root
        Root of the directory tree to analyze.
    virtualenv
        Virtual environment manager.
    """

    async def analyze(
        self, root: Path, *, update: bool = False
    ) -> list[Update]:
        """Analyze a tree and return needed Python frozen dependency updates.

        Parameters
        ----------
        root
            Root of the path to analyze.
        update
            If set to `True`, leave the update applied. This avoids having to
            run ``make update-deps`` twice, once to see if an update is needed
            and again to apply it properly.

        Returns
        -------
        list of Update
            Either an empty list (no updates needed) or a list with a single
            element (an update needed).

        Raises
        ------
        UncommittedChangesError
            Raised if the repository being analyzed has uncommitted changes
            and therefore cannot be checked for updates.
        subprocess.CalledProcessError
            Raised if running ``make update-deps`` failed.
        """
        for name in ("Makefile", "requirements/main.in"):
            if not (root / name).exists():
                return []
        repo = Repo(str(root))

        if repo.is_dirty():
            msg = "Working tree contains uncommitted changes"
            raise UncommittedChangesError(msg)

        try:
            subprocess.run(
                ["make", "update-deps"],
                cwd=str(root),
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            logging.exception(
                "make update-deps failed: %s%s", e.stdout, e.stderr
            )
            return []

        if not repo.is_dirty():
            return []

        if not update:
            repo.git.restore(".")
        return [PythonFrozenUpdate(path=root / "requirements", applied=update)]

    @property
    def name(self) -> str:
        return "python"



"""Base class for dependency updates.

Notes
-----
This is a bit more complicated than it should have to be to work around
https://github.com/python/mypy/issues/5374.  The mixins avoid the conflict
between an abstract base class and a dataclass in mypy's understanding of the
Python type system.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from pathlib import Path

__all__ = [
    "MethodMixin",
    "Update",
    "UpdateMixin",
]


class MethodMixin(ABC):
    """Add the abstract methods for an update."""

    @abstractmethod
    def apply(self) -> None:
        """Apply an update.

        Raises
        ------
        neophile.exceptions.DependencyNotFoundError
            Raised if the specified file doesn't contain a dependency of that
            name.
        """

    @abstractmethod
    def description(self) -> str:
        """Build a description of this update.

        Returns
        -------
        str
            Short text description of the update.
        """


@dataclass(order=True)
class UpdateMixin:
    """Add the base data elements for `Update`."""

    path: Path
    """The file that contains the dependency."""

    applied: bool
    """Whether the update has already been applied."""

    def to_dict(self) -> dict[str, str | bool]:
        """Convert the object to a dict.

        Returns
        -------
        dict of str or bool
            Dictionary representation of this object.

        Notes
        -----
        Required because ruamel.yaml cannot serialize `~pathlib.Path`.
        """
        result = asdict(self)
        result["path"] = str(result["path"])
        return result


class Update(UpdateMixin, MethodMixin):
    """Base class for a needed dependency version update."""



"""Process a set of repositories for updates."""

from __future__ import annotations

from pathlib import Path

from .analysis.base import BaseAnalyzer
from .config import Config
from .pr import PullRequester
from .repository import Repository
from .update.base import Update

__all__ = ["Processor"]


class Processor:
    """Process a set of repositories for updates.

    Parameters
    ----------
    config
        neophile configuration.
    analyzers
        Analyzers to run on the repositories.
    pull_requester
        Used to create pull requests.
    """

    def __init__(
        self,
        config: Config,
        analyzers: list[BaseAnalyzer],
        pull_requester: PullRequester,
    ) -> None:
        self._config = config
        self._analyzers = analyzers
        self._pull_requester = pull_requester

    async def analyze_checkout(self, path: Path) -> dict[str, list[Update]]:
        """Analyze a cloned repository without applying updates.

        Parameters
        ----------
        path
            Path to the cloned repository.

        Returns
        -------
        dict of Update
            Any updates found, organized by the analyzer that found the
            update.
        """
        results = {}
        for analyzer in self._analyzers:
            analysis = await analyzer.analyze(path)
            if analysis:
                results[analyzer.name] = analysis
        return results

    async def process_checkout(self, path: Path) -> None:
        """Check a cloned repository for updates.

        Creates pull requests as necessary if any needed updates are found.

        Parameters
        ----------
        path
            Path to the cloned repository.
        """
        repo = Repository(path)
        await self._process_one_repository(repo, path)

    async def update_checkout(self, path: Path) -> list[Update]:
        """Update a cloned repository.

        This does not switch branches. Updates are written to the current
        working tree.

        Parameters
        ----------
        path
            Path to the cloned repository.

        Returns
        -------
        list of Update
            All the updates that were applied.
        """
        all_updates = []
        for analyzer in self._analyzers:
            updates = await analyzer.update(path)
            all_updates.extend(updates)
        return all_updates

    async def _process_one_repository(
        self, repo: Repository, path: Path
    ) -> None:
        """Check a single repository for updates.

        Creates pull requests as necessary if any needed updates are found.

        Parameters
        ----------
        repo
            Cloned repository to check.
        path
            Path to the cloned repository.
        """
        repo.switch_branch()
        updates = await self.update_checkout(path)
        if updates:
            await self._pull_requester.make_pull_request(path, updates)
        repo.restore_branch()



"""Tests for the PullRequester class."""

from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
import respx
from git import PushInfo, Remote
from git.repo import Repo
from httpx import AsyncClient, Request, Response
from pydantic import SecretStr

from neophile.config import Config
from neophile.exceptions import PushError
from neophile.pr import CommitMessage, GitHubRepository, PullRequester
from neophile.repository import Repository
from neophile.update.pre_commit import PreCommitUpdate

from .support.github import mock_app_authenticate, mock_enable_auto_merge
from .util import setup_python_repo


@pytest.mark.asyncio
async def test_pr(
    tmp_path: Path,
    client: AsyncClient,
    respx_mock: respx.Router,
    github_key: str,
    mock_push: Mock,
) -> None:
    repo = setup_python_repo(tmp_path)
    Remote.create(repo, "origin", "https://github.com/foo/bar")
    config = Config(github_private_key=SecretStr(github_key))
    update = PreCommitUpdate(
        path=tmp_path / ".pre-commit-config.yaml",
        applied=False,
        repository="https://github.com/ambv/black",
        current="19.10b0",
        latest="23.3.0",
    )
    mock_app_authenticate(respx_mock, "foo/bar")
    respx_mock.get(f"https://api.github.com/users/{config.username}").mock(
        return_value=Response(200, json={"id": 123456}),
    )
    respx_mock.get("https://api.github.com/repos/foo/bar").mock(
        return_value=Response(200, json={"default_branch": "main"})
    )
    pattern = r"https://api.github.com/repos/foo/bar/pulls\?.*base=main.*"
    respx_mock.get(url__regex=pattern).mock(
        return_value=Response(200, json=[])
    )
    respx_mock.post("https://api.github.com/repos/foo/bar/pulls").mock(
        return_value=Response(201, json={"number": 1})
    )
    mock_enable_auto_merge(respx_mock, "foo", "bar", "1")

    repository = Repository(tmp_path)
    repository.switch_branch()
    update.apply()
    pr = PullRequester(config, client)
    await pr.make_pull_request(tmp_path, [update])

    assert mock_push.call_args_list == [
        call("u/neophile:u/neophile", force=True)
    ]
    assert not repo.is_dirty()
    assert repo.head.ref.name == "u/neophile"
    commit = repo.head.commit
    expected_email = f"123456+{config.username}@users.noreply.github.com"
    assert commit.author.name == config.username
    assert commit.author.email == expected_email
    assert commit.committer.name == config.username
    assert commit.committer.email == expected_email
    change = "Update ambv/black pre-commit hook from 19.10b0 to 23.3.0"
    assert commit.message == f"{CommitMessage.title}\n\n- {change}\n"
    assert "tmp-neophile" not in [r.name for r in repo.remotes]


@pytest.mark.asyncio
async def test_pr_push_failure(
    tmp_path: Path,
    client: AsyncClient,
    respx_mock: respx.Router,
    github_key: str,
) -> None:
    repo = setup_python_repo(tmp_path)
    Remote.create(repo, "origin", "https://github.com/foo/bar")
    config = Config(
        commit_email="someone@example.com",
        github_private_key=SecretStr(github_key),
    )
    update = PreCommitUpdate(
        path=tmp_path / ".pre-commit-config.yaml",
        applied=False,
        repository="https://github.com/ambv/black",
        current="19.10b0",
        latest="23.3.0",
    )
    remote = Mock(spec=Remote)
    push_error = PushInfo(
        PushInfo.ERROR, None, "", remote, summary="Some error"
    )
    mock_app_authenticate(respx_mock, "foo/bar")
    respx_mock.get("https://api.github.com/repos/foo/bar").mock(
        return_value=Response(200, json={"default_branch": "main"})
    )
    pattern = r"https://api.github.com/repos/foo/bar/pulls\?.*base=main.*"
    respx_mock.get(url__regex=pattern).mock(
        return_value=Response(200, json=[])
    )

    pr = PullRequester(config, client)
    with patch.object(Remote, "push") as mock:
        mock.return_value = [push_error]
        with pytest.raises(PushError) as excinfo:
            await pr.make_pull_request(tmp_path, [update])

    assert "Some error" in str(excinfo.value)


@pytest.mark.asyncio
async def test_pr_no_automerge(
    tmp_path: Path,
    client: AsyncClient,
    respx_mock: respx.Router,
    github_key: str,
    mock_push: Mock,
) -> None:
    repo = setup_python_repo(tmp_path)
    Remote.create(repo, "origin", "https://github.com/foo/bar")
    config = Config(
        commit_email="someone@example.com",
        github_private_key=SecretStr(github_key),
    )
    update = PreCommitUpdate(
        path=tmp_path / ".pre-commit-config.yaml",
        applied=False,
        repository="https://github.com/ambv/black",
        current="19.10b0",
        latest="23.3.0",
    )
    mock_app_authenticate(respx_mock, "foo/bar")
    respx_mock.get("https://api.github.com/repos/foo/bar").mock(
        return_value=Response(200, json={"default_branch": "main"})
    )
    pattern = r"https://api.github.com/repos/foo/bar/pulls\?.*base=main.*"
    respx_mock.get(url__regex=pattern).mock(
        return_value=Response(200, json=[])
    )
    respx_mock.post("https://api.github.com/repos/foo/bar/pulls").mock(
        return_value=Response(201, json={"number": 1})
    )
    mock_enable_auto_merge(respx_mock, "foo", "bar", "1", fail=True)

    repository = Repository(tmp_path)
    repository.switch_branch()
    update.apply()
    pr = PullRequester(config, client)
    await pr.make_pull_request(tmp_path, [update])

    assert mock_push.call_args_list == [
        call("u/neophile:u/neophile", force=True)
    ]
    assert not repo.is_dirty()
    assert repo.head.ref.name == "u/neophile"
    commit = repo.head.commit
    assert commit.author.name == config.username
    assert commit.author.email == "someone@example.com"
    assert commit.committer.name == config.username
    assert commit.committer.email == "someone@example.com"
    change = "Update ambv/black pre-commit hook from 19.10b0 to 23.3.0"
    assert commit.message == f"{CommitMessage.title}\n\n- {change}\n"
    assert "tmp-neophile" not in [r.name for r in repo.remotes]


@pytest.mark.asyncio
async def test_pr_update(
    tmp_path: Path,
    client: AsyncClient,
    respx_mock: respx.Router,
    github_key: str,
    mock_push: Mock,
) -> None:
    """Test updating an existing PR."""
    repo = setup_python_repo(tmp_path)
    Remote.create(repo, "origin", "https://github.com/foo/bar")
    config = Config(
        username="neophile[bot]",
        commit_email="otheremail@example.com",
        github_private_key=SecretStr(github_key),
    )
    update = PreCommitUpdate(
        path=tmp_path / ".pre-commit-config.yaml",
        applied=False,
        repository="https://github.com/ambv/black",
        current="19.10b0",
        latest="23.3.0",
    )
    updated_pr = False

    def check_pr_update(request: Request) -> Response:
        change = "Update ambv/black pre-commit hook from 19.10b0 to 23.3.0"
        assert json.loads(request.content) == {
            "title": CommitMessage.title,
            "body": f"- {change}\n",
        }

        nonlocal updated_pr
        updated_pr = True
        return Response(200)

    mock_app_authenticate(respx_mock, "foo/bar")
    respx_mock.get("https://api.github.com/repos/foo/bar").mock(
        return_value=Response(200, json={})
    )
    pattern = r"https://api.github.com/repos/foo/bar/pulls\?.*base=main.*"
    respx_mock.get(url__regex=pattern).mock(
        return_value=Response(200, json=[{"number": 1234}])
    )
    respx_mock.patch("https://api.github.com/repos/foo/bar/pulls/1234").mock(
        side_effect=check_pr_update
    )
    mock_enable_auto_merge(respx_mock, "foo", "bar", "1234")

    repository = Repository(tmp_path)
    repository.switch_branch()
    update.apply()
    pr = PullRequester(config, client)
    await pr.make_pull_request(tmp_path, [update])

    assert mock_push.call_args_list == [
        call("u/neophile:u/neophile", force=True)
    ]
    assert not repo.is_dirty()
    assert repo.head.ref.name == "u/neophile"
    commit = repo.head.commit
    assert commit.author.name == "neophile[bot]"
    assert commit.author.email == "otheremail@example.com"
    assert commit.committer.name == "neophile[bot]"
    assert commit.committer.email == "otheremail@example.com"


@pytest.mark.asyncio
async def test_get_authenticated_remote(
    tmp_path: Path, client: AsyncClient
) -> None:
    repo = Repo.init(str(tmp_path), initial_branch="main")
    pr = PullRequester(Config(), client)

    remote = Remote.create(repo, "origin", "https://github.com/foo/bar")
    url = pr._get_authenticated_remote(repo, "some-token")
    assert url == "https://neophile:some-token@github.com/foo/bar"

    remote.set_url("https://foo@github.com:8080/foo/bar")
    url = pr._get_authenticated_remote(repo, "some-token")
    assert url == "https://neophile:some-token@github.com:8080/foo/bar"

    remote.set_url("git@github.com:bar/foo")
    url = pr._get_authenticated_remote(repo, "some-token")
    assert url == "https://neophile:some-token@github.com/bar/foo"

    remote.set_url("ssh://git:blahblah@github.com/baz/stuff")
    url = pr._get_authenticated_remote(repo, "some-token")
    assert url == "https://neophile:some-token@github.com/baz/stuff"


@pytest.mark.asyncio
async def test_get_github_repo(tmp_path: Path, client: AsyncClient) -> None:
    repo = Repo.init(str(tmp_path), initial_branch="main")
    pr = PullRequester(Config(), client)

    remote = Remote.create(repo, "origin", "git@github.com:foo/bar.git")
    github_repo = pr._get_github_repo(repo)
    assert github_repo == GitHubRepository(owner="foo", repo="bar")

    remote.set_url("https://github.com/foo/bar.git")
    github_repo = pr._get_github_repo(repo)
    assert github_repo == GitHubRepository(owner="foo", repo="bar")

    remote.set_url("ssh://git@github.com/foo/bar")
    github_repo = pr._get_github_repo(repo)
    assert github_repo == GitHubRepository(owner="foo", repo="bar")



"""Wrapper around a Git repository."""

from __future__ import annotations

from pathlib import Path
from typing import Self

from git.repo import Repo

__all__ = ["Repository"]


class Repository:
    """Wrapper around a Git repository to add some convenience functions.

    Parameters
    ----------
    path
        Root path of the Git repository.
    """

    @classmethod
    def clone_or_update(cls, path: Path, url: str) -> Self:
        """Clone a repository or update an existing repository.

        Parameters
        ----------
        path
            Path to where the clone should be kept (and may already exist).
        url
            URL of the remote repository.

        Returns
        -------
        Repository
            Newly-created repository object.
        """
        if path.is_dir():
            repo = cls(path)
            repo.update()
            return repo

        Repo.clone_from(url, str(path))
        return cls(path)

    def __init__(self, path: Path) -> None:
        self._repo = Repo(str(path))
        self._branch = self._repo.head.ref

    def restore_branch(self) -> None:
        """Switch back to the branch before switch_branch was called.

        Also deletes the neophile branch to clean up.
        """
        self._branch.checkout()
        self._repo.delete_head("u/neophile", force=True)

    def switch_branch(self) -> None:
        """Switch to the neophile working branch.

        Notes
        -----
        Currently this unconditionally creates the branch and fails if it
        already exists.  Eventually this will be smarter about updating the
        neophile branch as appropriate.
        """
        branch = self._repo.create_head("u/neophile")
        branch.checkout()

    def update(self) -> None:
        """Update an existing checkout to its current upstream."""
        if "main" in self._repo.heads:
            self._repo.heads["main"].checkout()
        else:
            self._repo.heads["master"].checkout()
        if "u/neophile" in (h.name for h in self._repo.heads):
            self._repo.delete_head("u/neophile", force=True)
        self._repo.git.restore(".")
        self._repo.remotes.origin.pull(ff_only=True)



"""Tests for the Repository class."""

from __future__ import annotations

from pathlib import Path

from git.repo import Repo
from git.util import Actor

from neophile.repository import Repository


def create_repo(upstream_path: Path, checkout_path: Path) -> Repo:
    """Create an upstream and downstream repository.

    Parameters
    ----------
    upstream_path : `pathlib.Path`
        Path to the upstream bare repository.
    checkout_path : `pathlib.Path`
        Path to the downstream checkout.

    Returns
    -------
    repo : `git.Repo`
        The repository for the downstream checkout.
    """
    repo = Repo.init(str(checkout_path), initial_branch="main")
    Repo.init(str(upstream_path), bare=True, initial_branch="main")
    actor = Actor("Someone", "someone@example.com")

    (checkout_path / "foo").write_text("initial contents\n")
    repo.index.add("foo")
    repo.index.commit("Initial commit", author=actor, committer=actor)
    origin = repo.create_remote("origin", str(upstream_path))
    origin.push(all=True)
    repo.heads.main.set_tracking_branch(origin.refs.main)

    return repo


def test_clone_or_update(tmp_path: Path) -> None:
    one_path = tmp_path / "one"
    two_path = tmp_path / "two"
    upstream_path = tmp_path / "upstream"
    one_repo = create_repo(upstream_path, one_path)
    actor = Actor("Someone", "someone@example.com")

    Repository.clone_or_update(two_path, str(upstream_path))
    assert (two_path / "foo").read_text() == "initial contents\n"

    (one_path / "foo").write_text("new contents\n")
    one_repo.index.add("foo")
    one_repo.index.commit("New commit", author=actor, committer=actor)
    one_repo.remotes.origin.push()

    Repository.clone_or_update(two_path, str(upstream_path))
    assert (two_path / "foo").read_text() == "new contents\n"


def test_update_dirty_repo(tmp_path: Path) -> None:
    upstream_path = tmp_path / "upstream"
    checkout_path = tmp_path / "checkout"
    repo = create_repo(upstream_path, checkout_path)

    repository = Repository(checkout_path)
    repository.switch_branch()
    (checkout_path / "foo").write_text("other garbage\n")

    # After a clone or update, the branch should not exist and the changes
    # should be unwound.
    repository = Repository.clone_or_update(checkout_path, str(upstream_path))
    assert "u/neophile" not in (h.name for h in repo.heads)
    assert (checkout_path / "foo").read_text() == "initial contents\n"



"""Command-line interface."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

import click
from httpx import AsyncClient
from ruamel.yaml import YAML
from safir.asyncio import run_with_asyncio

from .factory import Factory
from .inventory.github import GitHubInventory

__all__ = [
    "analyze",
    "check",
    "github_inventory",
    "help",
    "main",
    "scan",
    "update",
]


def print_yaml(results: Any) -> None:
    """Print some results to stdout as YAML."""
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.dump(results, sys.stdout)


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(message="%(version)s")
def main() -> None:
    """Command-line interface for neophile."""


@main.command()
@click.argument("topic", default=None, required=False, nargs=1)
@click.pass_context
def help(ctx: click.Context, topic: str | None) -> None:
    """Show help for any command."""
    # The help command implementation is taken from
    # https://www.burgundywall.com/post/having-click-help-subcommand
    if topic:
        if topic in main.commands:
            ctx.info_name = topic
            click.echo(main.commands[topic].get_help(ctx))
        else:
            raise click.UsageError(f"Unknown help topic {topic}", ctx)
    else:
        if not ctx.parent:
            raise RuntimeError("help called without topic or parent")
        click.echo(ctx.parent.get_help())


@main.command()
@click.argument("types", type=click.Choice(["pre-commit", "python"]), nargs=-1)
@click.option(
    "--path",
    type=click.Path(path_type=Path),
    default=Path.cwd(),
    help="Path to analyze (default: current directory).",
)
@run_with_asyncio
async def analyze(types: list[str], *, path: Path) -> None:
    """Analyze a path for pending upgrades."""
    async with AsyncClient() as client:
        factory = Factory(client)
        processor = factory.create_processor(types=types if types else None)
        results = await processor.analyze_checkout(path)
        if results:
            data = {k: [u.to_dict() for u in v] for k, v in results.items()}
            print_yaml(data)


@main.command()
@click.argument("types", type=click.Choice(["pre-commit", "python"]), nargs=-1)
@click.option(
    "--path",
    type=click.Path(path_type=Path),
    default=Path.cwd(),
    help="Path to analyze (default: current directory).",
)
@run_with_asyncio
async def check(types: list[str], *, path: Path) -> None:
    """Check whether dependencies are up-to-date."""
    async with AsyncClient() as client:
        factory = Factory(client)
        processor = factory.create_processor(types=types if types else None)
        results = await processor.analyze_checkout(path)
        for dependency in results:
            sys.stderr.write(f"{dependency} dependencies out of date\n")
        if results:
            sys.exit(1)


@main.command()
@click.argument("owner", required=True)
@click.argument("repo", required=True)
@run_with_asyncio
async def github_inventory(owner: str, repo: str) -> None:
    """Inventory available GitHub tags."""
    async with AsyncClient() as client:
        inventory = GitHubInventory(client)
        result = await inventory.inventory(owner, repo)
        if result:
            sys.stdout.write(result + "\n")
        else:
            sys.stderr.write(f"No versions found in {owner}/{repo}")


@main.command()
@click.option(
    "--path",
    type=click.Path(path_type=Path),
    default=Path.cwd(),
    help="Path to scan (default: current directory).",
)
@run_with_asyncio
async def scan(path: Path) -> None:
    """Scan a path for versions."""
    async with AsyncClient() as client:
        factory = Factory(client)
        scanners = factory.create_all_scanners()
        results = {s.name: s.scan(path) for s in scanners}
        print_yaml({k: [u.to_dict() for u in v] for k, v in results.items()})


@main.command()
@click.argument("types", type=click.Choice(["pre-commit", "python"]), nargs=-1)
@click.option(
    "--path",
    type=click.Path(path_type=Path),
    default=Path.cwd(),
    help="Path to analyze (default: current directory).",
)
@click.option(
    "--pr/--no-pr", default=False, help="Generate a pull request of changes."
)
@run_with_asyncio
async def update(types: list[str], *, path: Path, pr: bool) -> None:
    """Update dependencies."""
    async with AsyncClient() as client:
        factory = Factory(client)
        processor = factory.create_processor(types=types if types else None)
        if pr:
            await processor.process_checkout(path)
        else:
            await processor.update_checkout(path)



"""Mocks for testing code interacting with GitHub."""

from __future__ import annotations

import json
import os
from collections.abc import Sequence
from copy import deepcopy
from pathlib import Path
from urllib.parse import urlparse

import respx
from gidgethub import QueryError
from httpx import Request, Response
from ruamel.yaml import YAML

from neophile.pr import _GRAPHQL_ENABLE_AUTO_MERGE, _GRAPHQL_PR_ID

__all__ = [
    "mock_app_authenticate",
    "mock_enable_auto_merge",
    "mock_github_tags",
    "mock_github_tags_from_precommit",
]


def mock_app_authenticate(respx_mock: respx.Router, slug: str) -> None:
    """Set up mocks for the GitHub API calls used for app authentication.

    Parameters
    ----------
    respx_mock
        Mock object for HTTP requests.
    slug
        Slug for the GitHub repository.
    """
    url = f"https://api.github.com/repos/{slug}/installation"
    respx_mock.get(url).mock(return_value=Response(200, json={"id": 1}))
    url = "https://api.github.com/app/installations/1/access_tokens"
    response = Response(200, json={"token": "ghs_" + os.urandom(16).hex()})
    respx_mock.post(url).mock(return_value=response)


def mock_enable_auto_merge(
    respx_mock: respx.Router,
    owner: str,
    repo: str,
    pr_number: str,
    *,
    fail: bool = False,
) -> None:
    """Set up mocks for the GitHub API call to enable auto-merge.

    Parameters
    ----------
    respx_mock
        Mock object for HTTP requests.
    owner
        Owner of the repository.
    repo
        Name of the repository.
    pr_number
        Number of the PR for which auto-merge will be set.
    fail
        Whether to fail the request for automerge
    """
    first = True

    def graphql(request: Request) -> Response:
        data = json.loads(request.content)
        nonlocal first
        if first:
            assert data == {
                "query": _GRAPHQL_PR_ID,
                "variables": {
                    "owner": owner,
                    "repo": repo,
                    "pr_number": int(pr_number),
                },
            }
            first = False
            data = {"data": {"repository": {"pullRequest": {"id": "some-id"}}}}
            return Response(200, json=data)
        else:
            assert data == {
                "query": _GRAPHQL_ENABLE_AUTO_MERGE,
                "variables": {"pr_id": "some-id"},
            }
            if fail:
                msg = (
                    "Pull request is not in the correct state to enable"
                    " auto-merge"
                )
                response = {"errors": [{"message": msg}]}
                raise QueryError(response)
            data = {"data": {"actor": {"login": "some-user"}}}
            return Response(200, json=data)

    url = "https://api.github.com/graphql"
    respx_mock.post(url).mock(side_effect=graphql)


def mock_github_tags(
    respx_mock: respx.Router, path: str, tags: Sequence[str]
) -> None:
    """Register a list of tags for a GitHub repository.

    Parameters
    ----------
    respx_mock
        Mock object for HTTP requests.
    path
        GitHub repository as :samp:`{owner}/{repo}`.
    tags
        List of tags to return for that repository.
    """
    data = [{"name": version} for version in tags]
    respx_mock.get(f"https://api.github.com/repos/{path}/tags").mock(
        return_value=Response(200, json=data)
    )


def mock_github_tags_from_precommit(
    respx_mock: respx.Router,
    pre_commit: Path,
    extra: dict[str, list[str]] | None = None,
) -> None:
    """Register all the tags found in a pre-commit file, plus extras.

    Ensure all the tags currently found in a pre-commit file will be returned
    by the mocks of GitHub API queries, and add the additional versions
    listed.

    Parameters
    ----------
    respx_mock
        Mock object for HTTP requests.
    pre_commit
        Path to the pre-commit file whose versions should be registered.
    extra
        Mapping of GitHub repository (as :samp:`{owner}/{path}`) to additional
        tags that should be returned for that repository.
    """
    versions = deepcopy(extra or {})
    yaml = YAML()
    with pre_commit.open("r") as f:
        pre_commit_data = yaml.load(f)
        for entry in pre_commit_data["repos"]:
            repo = urlparse(entry["repo"]).path.lstrip("/")
            if repo not in versions:
                versions[repo] = []
            versions[repo].append(entry["rev"])
    for repo, tags in versions.items():
        mock_github_tags(respx_mock, repo, tags)



"""Tests for the command-line interface."""

from __future__ import annotations

import json
import shutil
from pathlib import Path
from unittest.mock import Mock, call

import respx
from click.testing import CliRunner
from git import Remote
from git.repo import Repo
from git.util import Actor
from httpx import Request, Response
from ruamel.yaml import YAML

from neophile.cli import main
from neophile.pr import CommitMessage

from .support.github import (
    mock_app_authenticate,
    mock_enable_auto_merge,
    mock_github_tags,
    mock_github_tags_from_precommit,
)


def test_help() -> None:
    runner = CliRunner()

    result = runner.invoke(main, ["-h"])
    assert result.exit_code == 0
    assert "Commands:" in result.output

    result = runner.invoke(main, ["help"])
    assert result.exit_code == 0
    assert "Commands:" in result.output

    result = runner.invoke(main, ["help", "scan"])
    assert result.exit_code == 0
    assert "Commands:" not in result.output
    assert "Options:" in result.output

    result = runner.invoke(main, ["help", "unknown-command"])
    assert result.exit_code != 0
    assert "Unknown help topic unknown-command" in result.output


def test_analyze(tmp_path: Path, respx_mock: respx.Router) -> None:
    runner = CliRunner()
    src = Path(__file__).parent / "data" / "python" / ".pre-commit-config.yaml"
    dst = tmp_path / ".pre-commit-config.yaml"
    shutil.copy(src, dst)
    mock_github_tags_from_precommit(
        respx_mock, src, {"ambv/black": ["20.0.0"]}
    )

    result = runner.invoke(main, ["analyze", "--path", str(tmp_path)])
    assert result.exit_code == 0
    yaml = YAML()
    data = yaml.load(result.output)
    repository = "https://github.com/ambv/black"
    assert data["pre-commit"][0]["repository"] == repository
    assert data["pre-commit"][0]["current"] == "19.10b0"
    assert data["pre-commit"][0]["latest"] == "20.0.0"

    # Try again with no changes required.
    mock_github_tags_from_precommit(respx_mock, src)
    result = runner.invoke(main, ["analyze", "--path", str(tmp_path)])
    assert not result.output
    assert result.exit_code == 0


def test_check(tmp_path: Path, respx_mock: respx.Router) -> None:
    runner = CliRunner()
    src = Path(__file__).parent / "data" / "python" / ".pre-commit-config.yaml"
    dst = tmp_path / ".pre-commit-config.yaml"
    shutil.copy(src, dst)
    mock_github_tags_from_precommit(
        respx_mock, src, {"ambv/black": ["20.0.0"]}
    )

    result = runner.invoke(main, ["check", "--path", str(tmp_path)])
    assert result.exit_code == 1
    assert result.output == "pre-commit dependencies out of date\n"

    # Try again with no changes required.
    mock_github_tags_from_precommit(respx_mock, src)
    result = runner.invoke(main, ["check", "--path", str(tmp_path)])
    assert not result.output
    assert result.exit_code == 0


def test_github_inventory(respx_mock: respx.Router) -> None:
    mock_github_tags(respx_mock, "foo/bar", ["1.1.0", "1.2.0"])

    runner = CliRunner()
    result = runner.invoke(main, ["github-inventory", "foo", "bar"])
    assert result.exit_code == 0
    assert result.output == "1.2.0\n"


def test_scan() -> None:
    runner = CliRunner()

    path = Path(__file__).parent / "data" / "python"
    result = runner.invoke(main, ["scan", "--path", str(path)])
    assert result.exit_code == 0
    yaml = YAML()
    data = yaml.load(result.output)
    pre_commit_results = sorted(data["pre-commit"], key=lambda r: r["repo"])
    assert pre_commit_results[0]["version"] == "19.10b0"


def test_update(tmp_path: Path, respx_mock: respx.Router) -> None:
    runner = CliRunner()
    src = Path(__file__).parent / "data" / "python" / ".pre-commit-config.yaml"
    dst = tmp_path / ".pre-commit-config.yaml"
    shutil.copy(src, dst)
    yaml = YAML()
    mock_github_tags_from_precommit(
        respx_mock, src, {"ambv/black": ["20.0.0"]}
    )

    result = runner.invoke(
        main, ["update", "--path", str(tmp_path), "pre-commit"]
    )
    assert result.exit_code == 0
    data = yaml.load(dst)
    assert data["repos"][2]["rev"] == "20.0.0"


def test_update_pr(
    tmp_path: Path, respx_mock: respx.Router, github_key: str, mock_push: Mock
) -> None:
    runner = CliRunner()
    repo = Repo.init(str(tmp_path), initial_branch="main")
    Remote.create(repo, "origin", "https://github.com/foo/bar")
    src = Path(__file__).parent / "data" / "python" / ".pre-commit-config.yaml"
    dst = tmp_path / ".pre-commit-config.yaml"
    shutil.copy(src, dst)
    repo.index.add(str(dst))
    actor = Actor("Someone", "someone@example.com")
    repo.index.commit("Initial commit", author=actor, committer=actor)
    created_pr = False

    def check_pr_post(request: Request) -> Response:
        change = "Update ambv/black pre-commit hook from 19.10b0 to 20.0.0"
        assert json.loads(request.content) == {
            "title": CommitMessage.title,
            "body": f"- {change}\n",
            "head": "u/neophile",
            "base": "main",
            "maintainer_can_modify": True,
            "draft": False,
        }

        assert repo.head.ref.name == "u/neophile"
        yaml = YAML()
        data = yaml.load(dst)
        assert data["repos"][2]["rev"] == "20.0.0"
        commit = repo.head.commit
        assert commit.author.name == "neophile-square[bot]"
        assert commit.author.email == "someone@example.com"
        assert commit.message == f"{CommitMessage.title}\n\n- {change}\n"

        nonlocal created_pr
        created_pr = True
        return Response(201, json={"number": 42})

    mock_github_tags_from_precommit(
        respx_mock, src, {"ambv/black": ["20.0.0"]}
    )
    mock_app_authenticate(respx_mock, "foo/bar")
    respx_mock.get("https://api.github.com/repos/foo/bar").mock(
        return_value=Response(200, json={"default_branch": "main"})
    )
    pattern = r"https://api.github.com/repos/foo/bar/pulls\?.*"
    respx_mock.get(url__regex=pattern).mock(
        return_value=Response(200, json=[])
    )
    respx_mock.post("https://api.github.com/repos/foo/bar/pulls").mock(
        side_effect=check_pr_post
    )
    mock_enable_auto_merge(respx_mock, "foo", "bar", "42")

    result = runner.invoke(
        main,
        ["update", "--path", str(tmp_path), "--pr"],
        env={
            "NEOPHILE_COMMIT_EMAIL": "someone@example.com",
            "NEOPHILE_GITHUB_PRIVATE_KEY": github_key,
        },
    )
    assert result.exit_code == 0
    assert created_pr
    assert mock_push.call_args_list == [
        call("u/neophile:u/neophile", force=True)
    ]
    assert repo.head.ref.name == "main"



"""Python frozen dependency update."""

from __future__ import annotations

import logging
import subprocess
from dataclasses import dataclass
from pathlib import Path  # noqa: F401

from .base import Update

__all__ = ["PythonFrozenUpdate"]


@dataclass
class PythonFrozenUpdate(Update):
    """An update to Python frozen dependencies."""

    def apply(self) -> None:
        """Apply an update to frozen Python dependencies.

        Raises
        ------
        subprocess.CalledProcessError
            Raised if running ``make update-deps`` failed.
        """
        if self.applied:
            return
        rootdir = self.path.parent

        try:
            subprocess.run(
                ["make", "update-deps"],
                cwd=str(rootdir),
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            logging.exception(
                "make update-deps failed: %s%s", e.stdout, e.stderr
            )
            return

        self.applied = True

    def description(self) -> str:
        return "Update frozen Python dependencies"



"""Tests for the PreCommitAnalyzer class."""

from __future__ import annotations

from pathlib import Path

import pytest
import respx
from httpx import AsyncClient

from neophile.factory import Factory
from neophile.update.pre_commit import PreCommitUpdate

from ..support.github import mock_github_tags


@pytest.mark.asyncio
async def test_analyzer(client: AsyncClient, respx_mock: respx.Router) -> None:
    data_path = Path(__file__).parent.parent / "data" / "python"
    pre_commit_path = data_path / ".pre-commit-config.yaml"
    mock_github_tags(
        respx_mock,
        "pre-commit/pre-commit-hooks",
        ["v3.0.0", "v3.1.0", "v3.2.0"],
    )
    mock_github_tags(respx_mock, "timothycrosley/isort", ["4.3.21-2"])
    mock_github_tags(respx_mock, "ambv/black", ["20.0.0", "19.10b0"])
    mock_github_tags(respx_mock, "pycqa/flake8", ["3.7.0", "3.9.0"])

    factory = Factory(client)
    analyzer = factory.create_pre_commit_analyzer()
    results = await analyzer.analyze(data_path)

    assert results == [
        PreCommitUpdate(
            path=pre_commit_path,
            applied=False,
            repository="https://github.com/pre-commit/pre-commit-hooks",
            current="v3.1.0",
            latest="v3.2.0",
        ),
        PreCommitUpdate(
            path=pre_commit_path,
            applied=False,
            repository="https://github.com/ambv/black",
            current="19.10b0",
            latest="20.0.0",
        ),
        PreCommitUpdate(
            path=pre_commit_path,
            applied=False,
            repository="https://gitlab.com/pycqa/flake8",
            current="3.8.1",
            latest="3.9.0",
        ),
    ]



"""pre-commit hook dependency scanning."""

from __future__ import annotations

from pathlib import Path
from urllib.parse import urlparse

from ruamel.yaml import YAML

from ..models.dependencies import PreCommitDependency
from .base import BaseScanner

__all__ = ["PreCommitScanner"]


class PreCommitScanner(BaseScanner):
    """Scan a source tree for pre-commit hook version references."""

    def __init__(self) -> None:
        self._yaml = YAML()

    @property
    def name(self) -> str:
        return "pre-commit"

    def scan(self, root: Path) -> list[PreCommitDependency]:
        """Scan a source tree for pre-commit hook version references.

        Parameters
        ----------
        root
            Root of the source tree.

        Returns
        -------
        list of PreCommitDependency
            A list of all discovered pre-commit dependencies.
        """
        path = root / ".pre-commit-config.yaml"
        if not path.exists():
            return []

        results = []
        with path.open() as f:
            config = self._yaml.load(f)
        for hook in config.get("repos", []):
            path_components = urlparse(hook["repo"]).path[1:].split("/")
            dependency = PreCommitDependency(
                repository=hook["repo"],
                owner=path_components[0],
                repo=path_components[1],
                version=hook["rev"],
                path=path,
            )
            results.append(dependency)
        return results



"""Test for the Processor class."""

from __future__ import annotations

import json
import subprocess
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from unittest.mock import Mock, call, patch

import pytest
import respx
from git import PushInfo, Remote
from git.repo import Repo
from git.util import Actor
from httpx import AsyncClient, Request, Response
from pydantic import SecretStr
from ruamel.yaml import YAML

from neophile.config import Config
from neophile.factory import Factory
from neophile.pr import CommitMessage

from .support.github import (
    mock_app_authenticate,
    mock_enable_auto_merge,
    mock_github_tags_from_precommit,
)
from .util import setup_python_repo


def create_upstream_git_repository(repo: Repo, upstream_path: Path) -> None:
    """Create an upstream Git repository with Python files.

    Parameters
    ----------
    repo
        The repository to use as the contents of the upstream repository.
    upstream_path
        Where to put the upstream repository.
    """
    upstream_path.mkdir()
    Repo.init(str(upstream_path), bare=True, initial_branch="main")
    origin = repo.create_remote("origin", str(upstream_path))
    origin.push(all=True)


@contextmanager
def patch_clone_from(owner: str, repo: str, path: Path) -> Iterator[None]:
    """Patch :py:func:`git.Repo.clone_from` to check out a local repository.

    Parameters
    ----------
    owner
        GitHub repository owner to expect.
    repo
        GitHub repository name to expect.
    path
        File path to use as the true upstream location.
    """
    expected_url = f"https://github.com/{owner}/{repo}"

    def mock_clone_from(
        url: str,
        to_path: str,
        orig_clone: Callable[..., Repo] = Repo.clone_from,
        **kwargs: Any,
    ) -> Repo:
        assert url == expected_url
        repo = orig_clone(str(path), to_path)
        repo.remotes.origin.set_url(expected_url)
        return repo

    with patch.object(Repo, "clone_from", side_effect=mock_clone_from):
        yield


@pytest.mark.asyncio
async def test_processor(
    tmp_path: Path,
    client: AsyncClient,
    respx_mock: respx.Router,
    github_key: str,
) -> None:
    tmp_repo = setup_python_repo(tmp_path / "tmp")
    upstream_path = tmp_path / "upstream"
    create_upstream_git_repository(tmp_repo, upstream_path)
    with tmp_repo.remotes.origin.config_writer as cw:
        cw.set("url", "git@github.com:foo/bar")
    remote = Mock(spec=Remote)
    push_result = [PushInfo(PushInfo.NEW_HEAD, None, "", remote)]
    created_pr = False

    def check_pr_post(request: Request) -> Response:
        changes = [
            "Update frozen Python dependencies",
            "Update ambv/black pre-commit hook from 19.10b0 to 20.0.0",
        ]
        body = "- " + "\n- ".join(changes) + "\n"
        assert json.loads(request.content) == {
            "title": CommitMessage.title,
            "body": body,
            "head": "u/neophile",
            "base": "main",
            "maintainer_can_modify": True,
            "draft": False,
        }

        repo = Repo(str(tmp_path / "tmp"))
        assert repo.head.ref.name == "u/neophile"
        yaml = YAML()
        data = yaml.load(tmp_path / "tmp" / ".pre-commit-config.yaml")
        assert data["repos"][2]["rev"] == "20.0.0"
        commit = repo.head.commit
        assert commit.author.name == "neophile-square[bot]"
        assert commit.author.email == "someone@example.com"
        assert commit.message == f"{CommitMessage.title}\n\n{body}"

        nonlocal created_pr
        created_pr = True
        return Response(201, json={"number": 42})

    mock_github_tags_from_precommit(
        respx_mock,
        tmp_path / "tmp" / ".pre-commit-config.yaml",
        {"ambv/black": ["20.0.0"]},
    )
    mock_app_authenticate(respx_mock, "foo/bar")
    respx_mock.get("https://api.github.com/repos/foo/bar").mock(
        return_value=Response(200, json={"default_branch": "main"})
    )
    pattern = r"https://api.github.com/repos/foo/bar/pulls\?.*"
    respx_mock.get(url__regex=pattern).mock(
        return_value=Response(200, json=[])
    )
    respx_mock.post("https://api.github.com/repos/foo/bar/pulls").mock(
        side_effect=check_pr_post
    )
    mock_enable_auto_merge(respx_mock, "foo", "bar", "42")

    # Unfortunately, the mock_push fixture can't be used here because we
    # want to use git.Remote.push in create_upstream_git_repository.
    factory = Factory(client)
    factory._config = Config(
        commit_email="someone@example.com",
        github_private_key=SecretStr(github_key),
    )
    processor = factory.create_processor()
    with patch_clone_from("foo", "bar", upstream_path):
        with patch.object(Remote, "push") as mock_push:
            mock_push.return_value = push_result
            await processor.process_checkout(tmp_path / "tmp")

    assert mock_push.call_args_list == [
        call("u/neophile:u/neophile", force=True)
    ]
    assert created_pr
    assert not tmp_repo.is_dirty()
    assert tmp_repo.head.ref.name == "main"
    assert "u/neophile" not in [h.name for h in tmp_repo.heads]


@pytest.mark.asyncio
async def test_no_updates(
    tmp_path: Path, client: AsyncClient, respx_mock: respx.Router
) -> None:
    tmp_repo = setup_python_repo(tmp_path / "tmp")
    subprocess.run(
        ["make", "update-deps"], cwd=str(tmp_path / "tmp"), check=True
    )
    tmp_repo.index.add(str(tmp_path / "tmp" / "requirements"))
    actor = Actor("Someone", "someone@example.com")
    tmp_repo.index.commit("Update dependencies", author=actor, committer=actor)
    upstream_path = tmp_path / "upstream"
    create_upstream_git_repository(tmp_repo, upstream_path)
    mock_github_tags_from_precommit(
        respx_mock, tmp_path / "tmp" / ".pre-commit-config.yaml"
    )

    factory = Factory(client)
    processor = factory.create_processor()
    with patch_clone_from("foo", "bar", upstream_path):
        with patch.object(Remote, "push") as mock_push:
            await processor.process_checkout(tmp_path / "tmp")

    assert mock_push.call_count == 0
    assert not tmp_repo.is_dirty()
    assert tmp_repo.head.ref.name == "main"



"""Tests for the PythonAnalyzer class."""

from __future__ import annotations

import subprocess
from pathlib import Path

import pytest
from git.util import Actor
from httpx import AsyncClient

from neophile.exceptions import UncommittedChangesError
from neophile.factory import Factory
from neophile.update.python import PythonFrozenUpdate

from ..util import setup_python_repo


@pytest.mark.asyncio
async def test_analyzer(tmp_path: Path, client: AsyncClient) -> None:
    repo = setup_python_repo(tmp_path)
    actor = Actor("Someone", "someone@example.com")

    factory = Factory(client)
    analyzer = factory.create_python_analyzer()
    results = await analyzer.analyze(tmp_path)

    assert results == [
        PythonFrozenUpdate(path=tmp_path / "requirements", applied=False)
    ]

    # Ensure that the tree is restored to the previous contents.
    assert not repo.is_dirty()

    # If the repo is dirty, analysis will fail.
    subprocess.run(["make", "update-deps"], cwd=str(tmp_path), check=True)
    assert repo.is_dirty()
    factory = Factory(client)
    analyzer = factory.create_python_analyzer()
    with pytest.raises(UncommittedChangesError):
        results = await analyzer.analyze(tmp_path)

    # Commit the changed dependencies and remove the pre-commit configuration
    # file.  Analysis should now return no changes.
    repo.index.add(str(tmp_path / "requirements"))
    repo.index.commit("Update dependencies", author=actor, committer=actor)
    factory = Factory(client)
    analyzer = factory.create_python_analyzer()
    results = await analyzer.analyze(tmp_path)
    assert results == []


@pytest.mark.asyncio
async def test_analyzer_update(tmp_path: Path, client: AsyncClient) -> None:
    repo = setup_python_repo(tmp_path)

    factory = Factory(client)
    analyzer = factory.create_python_analyzer()
    results = await analyzer.update(tmp_path)

    assert results == [
        PythonFrozenUpdate(path=tmp_path / "requirements", applied=True)
    ]
    assert repo.is_dirty()



"""Base class for an analysis step."""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path

from ..update.base import Update

__all__ = ["BaseAnalyzer"]


class BaseAnalyzer(ABC):
    """Base class for an analysis step."""

    @abstractmethod
    async def analyze(
        self, root: Path, *, update: bool = False
    ) -> list[Update]:
        """Analyze a tree and return a list of needed changes.

        Parameters
        ----------
        root
            Root of the path to analyze.
        update
            If set to `True`, leave the update applied if this is more
            efficient. Used by analyzers like the Python frozen dependency
            analyzer that have to do work and apply the update to see if any
            update is necessary. They can then mark the returned update as
            already applied and not have to run it twice.

        Returns
        -------
        list of Update
            List of updates found.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the analyzer type.

        String representing the type of analyzer this is. Used for reporting
        results accumulated from a bunch of analyzers.
        """

    async def update(self, root: Path) -> list[Update]:
        """Analyze a tree and apply updates.

        Returns
        -------
        list of Update
            List of updates applied.
        """
        updates = await self.analyze(root, update=True)
        for update in updates:
            update.apply()
        return updates



"""Inventory of available GitHub tags."""

from __future__ import annotations

import logging

from gidgethub import GitHubException
from gidgethub.httpx import GitHubAPI
from httpx import AsyncClient, HTTPError

from .version import PackagingVersion, ParsedVersion, SemanticVersion

__all__ = ["GitHubInventory"]


class GitHubInventory:
    """Inventory available tags of a GitHub repository.

    Parameters
    ----------
    http_client
        HTTP client to use for requests.
    """

    def __init__(self, http_client: AsyncClient) -> None:
        self._github = GitHubAPI(http_client, "lsst-sqre/neophile")

    async def inventory(
        self, owner: str, repo: str, *, semantic: bool = False
    ) -> str | None:
        """Return the latest tag of a GitHub repository.

        Parameters
        ----------
        owner
            Owner of the repository.
        repo
            Name of the repository.
        semantic
            If set to true, only semantic versions will be considered and the
            latest version will be determined by semantic version sorting
            instead of `packaging.version.Version`.

        Returns
        -------
        str or None
            The latest tag in sorted order. Tags that parse as valid versions
            sort before tags that do not, which should normally produce the
            correct results when version tags are mixed with other tags. If no
            valid tags are found or the repository doesn't exist, returns
            `None`.
        """
        logging.info(f"Inventorying GitHub repo {owner}/{repo}")
        if semantic:
            cls: type[ParsedVersion] = SemanticVersion
        else:
            cls = PackagingVersion

        try:
            tags = self._github.getiter(
                "/repos{/owner}{/repo}/tags",
                url_vars={"owner": owner, "repo": repo},
            )
            versions = [
                cls.from_str(tag["name"])
                async for tag in tags
                if cls.is_valid(tag["name"])
            ]
        except (GitHubException, HTTPError) as e:
            error = type(e).__name__
            if str(e):
                error += f": {e!s}"
            msg = f"Unable to inventory GitHub repo {owner}/{repo}: {error}"
            logging.exception(msg)
            return None

        if versions:
            return str(max(versions))
        else:
            msg = f"No valid versions for GitHub repo {owner}/{repo}"
            logging.warning(msg)
            return None



"""The neophile service."""

__all__ = ["__version__"]

from importlib.metadata import PackageNotFoundError, version

__version__: str
"""The application version string (PEP 440 / SemVer compatible)."""

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"



"""Tests for the PreCommitUpdate class."""

from __future__ import annotations

import shutil
from pathlib import Path

import pytest
from ruamel.yaml import YAML

from neophile.exceptions import DependencyNotFoundError
from neophile.update.pre_commit import PreCommitUpdate


def test_update(tmp_path: Path) -> None:
    source_path = (
        Path(__file__).parent.parent
        / "data"
        / "python"
        / ".pre-commit-config.yaml"
    )
    config_path = tmp_path / ".pre-commit-config.yaml"
    shutil.copy(str(source_path), str(config_path))

    update = PreCommitUpdate(
        path=config_path,
        applied=False,
        repository="https://github.com/pre-commit/pre-commit-hooks",
        current="v3.1.0",
        latest="v3.1.1",
    )
    update.apply()

    description = (
        "Update pre-commit/pre-commit-hooks pre-commit hook from v3.1.0 to"
        " v3.1.1"
    )
    assert update.description() == description

    yaml = YAML()
    expected = yaml.load(source_path)
    expected["repos"][0]["rev"] = "v3.1.1"
    assert yaml.load(config_path) == expected


def test_update_not_found() -> None:
    source_path = (
        Path(__file__).parent.parent
        / "data"
        / "python"
        / ".pre-commit-config.yaml"
    )

    update = PreCommitUpdate(
        path=source_path,
        applied=False,
        repository="https://github.com/foo/bar",
        current="1.0.0",
        latest="1.2.0",
    )
    with pytest.raises(DependencyNotFoundError):
        update.apply()
    assert not update.applied

    # Test that if the update is already applied, we won't try again, by
    # running apply and showing that it doesn't thrown an exception.
    update.applied = True
    update.apply()



"""Utility functions for tests."""

from __future__ import annotations

import shutil
from collections.abc import Mapping
from io import StringIO
from pathlib import Path
from typing import Any

from git.repo import Repo
from git.util import Actor
from ruamel.yaml import YAML

__all__ = [
    "dict_to_yaml",
    "setup_python_repo",
]


def dict_to_yaml(data: Mapping[str, Any]) -> str:
    """Convert any mapping to YAML serialized as a string.

    Parameters
    ----------
    data
        Data to convert.

    Returns
    -------
    str
        Data serialized as YAML.
    """
    yaml = YAML()
    output = StringIO()
    yaml.dump(data, output)
    return output.getvalue()


def setup_python_repo(tmp_path: Path) -> Repo:
    """Set up a test repository with the Python test files.

    Parameters
    ----------
    tmp_path
        The directory in which to create the repository.

    Returns
    -------
    Repo
        Repository object.
    """
    data_path = Path(__file__).parent / "data" / "python"
    shutil.copytree(str(data_path), str(tmp_path), dirs_exist_ok=True)
    repo = Repo.init(str(tmp_path), initial_branch="main")
    repo.index.add(
        [
            str(tmp_path / ".pre-commit-config.yaml"),
            str(tmp_path / "Makefile"),
            str(tmp_path / "requirements"),
        ]
    )
    actor = Actor("Someone", "someone@example.com")
    repo.index.commit("Initial commit", author=actor, committer=actor)
    return repo



"""Factory for neophile components."""

from __future__ import annotations

from httpx import AsyncClient

from .analysis.base import BaseAnalyzer
from .analysis.pre_commit import PreCommitAnalyzer
from .analysis.python import PythonAnalyzer
from .config import Config
from .inventory.github import GitHubInventory
from .pr import PullRequester
from .processor import Processor
from .scanner.base import BaseScanner
from .scanner.pre_commit import PreCommitScanner

__all__ = ["Factory"]


class Factory:
    """Factory to create neophile components.

    Parameters
    ----------
    http_client
        HTTP client to use for requests.
    """

    def __init__(self, http_client: AsyncClient) -> None:
        self._http_client = http_client
        self._config = Config()

    def create_analyzers(
        self, types: list[str] | None = None
    ) -> list[BaseAnalyzer]:
        """Create all analyzers.

        Parameters
        ----------
        types
            If given, only create analyzers of the given types.

        Returns
        -------
        list of BaseAnalyzer
            List of analyzers.

        Notes
        -----
        The Python analyzer requires a clean Git tree in order to determine if
        any changes were necessary, and therefore must run first if the
        analyzers are run in update mode (which means they will make changes
        to the working tree).
        """
        if not types:
            types = ["python", "pre-commit"]

        analyzers: list[BaseAnalyzer] = []
        if "python" in types:
            analyzers.append(self.create_python_analyzer())
        if "pre-commit" in types:
            analyzers.append(self.create_pre_commit_analyzer())
        return analyzers

    def create_all_scanners(self) -> list[BaseScanner]:
        """Create all scanners.

        Returns
        -------
        list of BaseScanner
            List of all available scanners.
        """
        return [PreCommitScanner()]

    def create_pre_commit_analyzer(self) -> PreCommitAnalyzer:
        """Create a new pre-commit hook analyzer.

        Returns
        -------
        PreCommitAnalyzer
            New analyzer.
        """
        scanner = PreCommitScanner()
        inventory = GitHubInventory(self._http_client)
        return PreCommitAnalyzer(scanner, inventory)

    def create_python_analyzer(self) -> PythonAnalyzer:
        """Create a new Python frozen dependency analyzer.

        Returns
        -------
        PythonAnalyzer
            New analyzer.
        """
        return PythonAnalyzer()

    def create_processor(self, types: list[str] | None = None) -> Processor:
        """Create a new repository processor.

        Parameters
        ----------
        types
            If given, only process dependencies of the given types.

        Returns
        -------
        Processor
            New processor.
        """
        return Processor(
            self._config,
            self.create_analyzers(types),
            self.create_pull_requester(),
        )

    def create_pull_requester(self) -> PullRequester:
        """Create a new pull requester.

        Returns
        -------
        PullRequester
            New pull requester.
        """
        return PullRequester(self._config, self._http_client)



"""Tests for the PreCommitScanner class."""

from __future__ import annotations

from pathlib import Path

from neophile.models.dependencies import PreCommitDependency
from neophile.scanner.pre_commit import PreCommitScanner


def test_scanner() -> None:
    data_path = Path(__file__).parent.parent / "data" / "python"
    scanner = PreCommitScanner()
    results = scanner.scan(data_path)

    assert results == [
        PreCommitDependency(
            repository="https://github.com/pre-commit/pre-commit-hooks",
            owner="pre-commit",
            repo="pre-commit-hooks",
            version="v3.1.0",
            path=data_path / ".pre-commit-config.yaml",
        ),
        PreCommitDependency(
            repository="https://github.com/timothycrosley/isort",
            owner="timothycrosley",
            repo="isort",
            version="4.3.21-2",
            path=data_path / ".pre-commit-config.yaml",
        ),
        PreCommitDependency(
            repository="https://github.com/ambv/black",
            owner="ambv",
            repo="black",
            version="19.10b0",
            path=data_path / ".pre-commit-config.yaml",
        ),
        PreCommitDependency(
            repository="https://gitlab.com/pycqa/flake8",
            owner="pycqa",
            repo="flake8",
            version="3.8.1",
            path=data_path / ".pre-commit-config.yaml",
        ),
    ]



"""Analysis of a repository for needed pre-commit hook updates."""

from __future__ import annotations

from pathlib import Path

from ..inventory.github import GitHubInventory
from ..scanner.pre_commit import PreCommitScanner
from ..update.base import Update
from ..update.pre_commit import PreCommitUpdate
from .base import BaseAnalyzer

__all__ = ["PreCommitAnalyzer"]


class PreCommitAnalyzer(BaseAnalyzer):
    """Analyze a tree for needed pre-commit hook updates.

    Parameters
    ----------
    scanner
        Scanner for pre-commit hook dependencies.
    inventory
        Inventory for GitHub tags.
    """

    def __init__(
        self, scanner: PreCommitScanner, inventory: GitHubInventory
    ) -> None:
        self._scanner = scanner
        self._inventory = inventory

    async def analyze(
        self, root: Path, *, update: bool = False
    ) -> list[Update]:
        """Analyze a tree and return needed pre-commit hook changes.

        Parameters
        ----------
        root
            Root of the path to analyze.
        update
            Ignored for this analyzer.

        Returns
        -------
        list of Update
            List of needed updates.
        """
        dependencies = self._scanner.scan(root)

        results: list[Update] = []
        for dependency in dependencies:
            latest = await self._inventory.inventory(
                dependency.owner, dependency.repo
            )
            if latest is not None and latest != dependency.version:
                pre_commit_update = PreCommitUpdate(
                    path=dependency.path,
                    applied=False,
                    repository=dependency.repository,
                    current=dependency.version,
                    latest=latest,
                )
                results.append(pre_commit_update)

        return results

    @property
    def name(self) -> str:
        return "pre-commit"



"""Version representation for inventories."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Self

from packaging import version
from semver.version import Version

__all__ = [
    "PackagingVersion",
    "ParsedVersion",
    "SemanticVersion",
]


@dataclass(frozen=True, order=True)
class ParsedVersion(metaclass=ABCMeta):
    """Abstract base class for versions.

    We use two separate version implementations, one based on
    `packaging.version.Version` and one based on `semver.version.Version`.
    This class defines the common interface.
    """

    @classmethod
    @abstractmethod
    def from_str(cls, string: str) -> Self:
        """Parse a string into a version.

        Parameters
        ----------
        string
            Version as a string.

        Returns
        -------
        ParsedVersion
            Parsed version.
        """

    @staticmethod
    @abstractmethod
    def is_valid(string: str) -> bool:
        """Return whether a version string is a valid version.

        Parameters
        ----------
        string
            Version as a string.

        Returns
        -------
        bool
            Whether it is valid.
        """

    @abstractmethod
    def __str__(self) -> str:
        """Original form of the version."""


@dataclass(frozen=True, order=True)
class PackagingVersion(ParsedVersion):
    """Represents a version string using `packaging.version.Version`."""

    parsed_version: version.Version
    """The canonicalized, parsed version, for sorting.

    Notes
    -----
    This field must be first because it's the field we want to sort on and
    dataclass ordering is done as if the dataclass were a tuple, via ordering
    on each element of the tuple in sequence.
    """

    version: str
    """The raw version string."""

    @classmethod
    def from_str(cls, string: str) -> Self:
        """Parse a string into a `~packaging.version.Version`.

        Parameters
        ----------
        string
            Version as a string.

        Returns
        -------
        PackagingVersion
            Parsed version.
        """
        parsed_version = version.parse(string)
        return cls(parsed_version=parsed_version, version=string)

    @staticmethod
    def is_valid(string: str) -> bool:
        """Return whether a version is valid.

        Parameters
        ----------
        string
            Version as a string.

        Returns
        -------
        bool
            Always returns `True` since all versions are valid for this
            implementation (some will parse to a legacy version).
        """
        try:
            version.parse(string)
        except version.InvalidVersion:
            return False
        else:
            return True

    def __str__(self) -> str:
        return self.version


@dataclass(frozen=True, order=True)
class SemanticVersion(ParsedVersion):
    """Represents a semantic version string."""

    parsed_version: Version
    """The parsed version of it, for sorting.

    Notes
    -----
    This field must be first because it's the field we want to sort on and
    dataclass ordering is done as if the dataclass were a tuple, via ordering
    on each element of the tuple in sequence.
    """

    version: str
    """The raw version string, which may start with a v."""

    @classmethod
    def from_str(cls, string: str) -> Self:
        """Parse a string into a `SemanticVersion`.

        Parameters
        ----------
        string
            Version as a string.

        Returns
        -------
        SemanticVersion
            Parsed version.
        """
        version = string[1:] if string.startswith("v") else string
        return cls(version=string, parsed_version=Version.parse(version))

    @staticmethod
    def is_valid(string: str) -> bool:
        version = string[1:] if string.startswith("v") else string
        return Version.is_valid(version)

    def __str__(self) -> str:
        return self.version



"""pre-commit dependency update."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path  # noqa: F401
from urllib.parse import urlparse

from ruamel.yaml import YAML

from ..exceptions import DependencyNotFoundError
from .base import Update

__all__ = ["PreCommitUpdate"]


@dataclass(order=True)
class PreCommitUpdate(Update):
    """An update to a Helm chart dependency."""

    repository: str
    """The URL of the GitHub repository providing this pre-commit hook."""

    current: str
    """The current version."""

    latest: str
    """The latest available version."""

    def apply(self) -> None:
        """Apply an update to a Helm chart.

        Raises
        ------
        DependencyNotFoundError
            Raised if the specified file doesn't contain a dependency of that
            name.
        """
        if self.applied:
            return

        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        data = yaml.load(self.path)

        found = False
        for hook in data.get("repos", []):
            if hook["repo"] == self.repository:
                hook["rev"] = self.latest
                found = True
        if not found:
            raise DependencyNotFoundError(
                f"Cannot find dependency for {self.repository} in {self.path}"
            )

        with self.path.open("w") as f:
            yaml.dump(data, f)

        self.applied = True

    def description(self) -> str:
        short_repo = urlparse(self.repository).path[1:]
        return (
            f"Update {short_repo} pre-commit hook from {self.current} to"
            f" {self.latest}"
        )



"""Representations of dependencies."""

from __future__ import annotations

from dataclasses import asdict, dataclass
from pathlib import Path

__all__ = [
    "Dependency",
    "HelmDependency",
    "KustomizeDependency",
    "PreCommitDependency",
]


@dataclass(frozen=True, order=True)
class Dependency:
    """Base class for a dependency returned by a scanner."""

    path: Path
    """The file that contains the dependency declaration."""

    def to_dict(self) -> dict[str, str]:
        """Convert the object to a dict.

        Notes
        -----
        Required because ruamel.yaml cannot serialize `~pathlib.Path`.
        """
        result = asdict(self)
        result["path"] = str(result["path"])
        return result


@dataclass(frozen=True, order=True)
class HelmDependency(Dependency):
    """Represents a single Helm dependency."""

    version: str
    """The version of the dependency (may be a match pattern)."""

    name: str
    """The name of the external dependency."""

    repository: str
    """The name of the chart repository containing the dependency."""


@dataclass(frozen=True, order=True)
class KustomizeDependency(Dependency):
    """Represents a single Kustomize dependency."""

    url: str
    """The full URL of the dependency."""

    owner: str
    """The owner of the referenced GitHub repository."""

    repo: str
    """The name of the referenced GitHub repository."""

    version: str
    """The version of the dependency."""


@dataclass(frozen=True, order=True)
class PreCommitDependency(Dependency):
    """Represents a single pre-commit dependency."""

    repository: str
    """The URL of the GitHub repository providing this pre-commit hook."""

    owner: str
    """The GitHub repository owner of the pre-commit hook."""

    repo: str
    """The GitHub repository name of the pre-commit hook."""

    version: str
    """The version of the dependency (may be a match pattern)."""
